module eposcorr_io

  use param  
  use dal
  use eposcorr_def

  implicit none

  private

  public read_srclist, read_catset, modify_list, write_image

contains

  !--------------------------------------------------------------------
  function read_srclist( filename, n_sources, ra_pnt, dec_pnt, &
           xra, xdec, xerr, xext) result( list_ptr)
  !--------------------------------------------------------------------

    type(Coordinate_type), dimension(:), pointer :: list_ptr
    character(*), intent(in) :: filename, xra, xdec, xerr, xext
    integer(kind=int32) , intent (out) :: n_sources
    real(kind=double) , intent (out) :: ra_pnt, dec_pnt

    real(kind=double),dimension(:),pointer :: ra, dec
    type(DataSetT) :: set
    type(TableT)  :: source_tab
    type(ColumnT) :: raCol
    type(ColumnT) :: decCol
    type(ColumnT) :: errCol
    integer(kind=int32) :: i, nrows

    real(kind=single),dimension(:),pointer :: err

    set = dataSet(filename,READ)

    if(  hasAttribute(set, "DEC_PNT") ) then
       dec_pnt  = real64Attribute(set, "DEC_PNT")
       if(  hasAttribute(set, "RA_PNT") ) then
          ra_pnt  = real64Attribute(set, "RA_PNT")
       else
          ra_pnt  = 1000.0
       end if
    else
       ra_pnt   = 0.0
       dec_pnt  = nodec_pnt
    end if

    source_tab = table( set,xext)

    ! get number of rows in this table
    nrows = numberOfRows(source_tab)
    if( nrows <= 0 ) then
       call release(source_tab) ! probably superfluous
       call release(set)        ! probably superfluous
       call error('emptySrcList','Number of rows in source list is 0')
    end if

    raCol = column(source_tab, xra, READ)
    decCol = column(source_tab, xdec, READ)

    ra => real64Data( raCol )
    dec => real64Data( decCol )

    ! allocate position list
    allocate(list_ptr(nrows))
    n_sources = nrows

    ! Copy the fits data into the structure
    !
    if (index(units(raCol),"deg")>0) then
      do i=1, n_sources
        list_ptr(i)%ra = ra(i-1)*deg2rad
      end do
    else if (index(units(raCol),"rad")>0) then
      do i=1, n_sources
        list_ptr(i)%ra = ra(i-1)
      end do
    else 
      call warning("NoUnit","No RA unit; deg assumed")
      do i=1, n_sources
        list_ptr(i)%ra = ra(i-1)*deg2rad
      end do
    end if 

    if (index(units(decCol),"deg")>0) then
      do i=1, n_sources
        list_ptr(i)%dec = dec(i-1)*deg2rad 
      end do
    else if (index(units(decCol),"rad")>0) then
      do i=1, n_sources
        list_ptr(i)%dec = dec(i-1) 
      end do
    else 
      call warning("NoUnit","No DEC unit; deg assumed")
      do i=1, n_sources
        list_ptr(i)%dec = dec(i-1)*deg2rad 
      end do
    end if 

    if (hasColumn(source_tab,xerr)) then

      errCol = column(source_tab, xerr, READ)
      err => real32Data( errCol )

      if (index(units(errCol),"sec")>0) then
        do i=1, n_sources
          list_ptr(i)%ra_err = sqrt(0.5)*err(i-1)
          list_ptr(i)%dec_err = sqrt(0.5)*err(i-1)
          list_ptr(i)%error = sqrt(0.5)*err(i-1)
        end do
      else if (index(units(errCol),"deg")>0) then
        do i=1, n_sources
          list_ptr(i)%ra_err = sqrt(0.5)*err(i-1)*3600d0
          list_ptr(i)%dec_err = sqrt(0.5)*err(i-1)*3600d0
          list_ptr(i)%error = sqrt(0.5)*err(i-1)*3600d0
        end do
      else if (index(units(errCol),"rad")>0) then
        do i=1, n_sources
          list_ptr(i)%ra_err = sqrt(0.5)*err(i-1)*3600d0/deg2rad
          list_ptr(i)%dec_err = sqrt(0.5)*err(i-1)*3600d0/deg2rad
          list_ptr(i)%error = sqrt(0.5)*err(i-1)*3600d0/deg2rad
        end do
      else 
        call warning("NoUnit","No error unit; arcsec assumed")
        do i=1, n_sources
          list_ptr(i)%ra_err = sqrt(0.5)*err(i-1)
          list_ptr(i)%dec_err = sqrt(0.5)*err(i-1)
          list_ptr(i)%error = sqrt(0.5)*err(i-1)
        end do
      end if 

      call release( errCol )

    else 
      do i=1, n_sources
        list_ptr(i)%ra_err = sqrt(0.5)
        list_ptr(i)%dec_err = sqrt(0.5)
        list_ptr(i)%error = sqrt(0.5)
      end do
    end if
  
    call release( raCol )
    call release( decCol )
    call release(source_tab)
    call release(set)


  end function read_srclist

  !--------------------------------------------------------------------
  function read_catset( filename, n_sources, bmin, bmax, usebmagn, & 
           rmin, rmax, usermagn, cra, cdec, cerr, cbmg, crmg, cext) &
           result( list_ptr)
  !--------------------------------------------------------------------

    type(Coordinate_type), dimension(:), pointer :: list_ptr

    character(*), intent(in) :: filename, cra, cdec, cerr, cext, cbmg,crmg
    integer(kind=int32) , intent (out) :: n_sources
    real(kind=single) , intent (in) :: bmin, bmax, rmin, rmax
    logical, intent (in out) :: usebmagn, usermagn 

    type(DataSetT) :: set
    type(TableT)  :: source_tab
    type(ColumnT) :: raCol
    type(ColumnT) :: decCol
    type(ColumnT) :: errCol
    type(ColumnT) :: bmagnCol, rmagnCol
    integer(kind=int32) :: i, nrows, nnew
    real(kind=double),dimension(:),pointer :: ra, dec
    real(kind=single),dimension(:),pointer :: err, bmagn, rmagn

    set = dataSet(filename,READ)
    source_tab = table( set, cext)

    ! get number of rows in this table
    nrows = numberOfRows(source_tab)
    if( nrows <= 0 ) then
       call release(source_tab) ! probably superfluous
       call release(set)        ! probably superfluous
       call error('emptyCatList','Number of rows in optical source list is 0')
    end if

    raCol = column(source_tab, cra,READ)
    decCol = column(source_tab, cdec, READ)

    ra => real64Data( raCol )
    dec => real64Data( decCol )

    if( usebmagn .AND. hasColumn(source_tab,cbmg)) then
       bmagnCol = column(source_tab,cbmg,READ)
       bmagn => real32Data( bmagnCol )
    else
       if( usebmagn )then
          call warning('noBmagnColumn','Column Bmagn does not exist.')
       endif
       allocate(bmagn(0:nrows))
       usebmagn = .FALSE.
    end if

    if( usermagn .AND. hasColumn(source_tab,crmg)) then
       rmagnCol = column(source_tab,crmg,READ)
       rmagn => real32Data( rmagnCol )
    else
       if( usermagn )then
          call warning('noBmagnColumn','Column Rmagn does not exist.')
       endif
       allocate(rmagn(0:nrows))
       usermagn = .FALSE.
    end if

    ! allocate position list
    allocate(list_ptr(nrows))
    n_sources = nrows

    ! Copy the fits data into the structure
    !

    if (index(units(raCol),"deg")>0) then
      nnew=0
      do i=1, n_sources
       if(((usermagn .AND. rmagn(i-1) >= rmin .AND. rmagn(i-1) <= rmax) .OR. &
            .NOT.usermagn) .AND. &
          ((usebmagn .AND. bmagn(i-1) >= bmin .AND. bmagn(i-1) <= bmax) .OR. &
               .NOT.usebmagn)) then
          nnew = nnew + 1
          list_ptr(nnew)%ra = ra(i-1)*deg2rad   
       end if
      end do
    else if (index(units(raCol),"rad")>0) then
      nnew=0
      do i=1, n_sources
       if(((usermagn .AND. rmagn(i-1) >= rmin .AND. rmagn(i-1) <= rmax) .OR. &
            .NOT.usermagn) .AND. &
          ((usebmagn .AND. bmagn(i-1) >= bmin .AND. bmagn(i-1) <= bmax) .OR. &
               .NOT.usebmagn)) then
          nnew = nnew + 1
          list_ptr(nnew)%ra = ra(i-1)
       end if
      end do
    else
      call warning("NoUnit","No RA unit; deg assumed")
      nnew=0
      do i=1, n_sources
       if(((usermagn .AND. rmagn(i-1) >= rmin .AND. rmagn(i-1) <= rmax) .OR. &
            .NOT.usermagn) .AND. &
          ((usebmagn .AND. bmagn(i-1) >= bmin .AND. bmagn(i-1) <= bmax) .OR. &
               .NOT.usebmagn)) then
          nnew = nnew + 1
          list_ptr(nnew)%ra = ra(i-1)*deg2rad   
       end if
      end do
    end if

    if (index(units(decCol),"deg")>0) then
      nnew=0
      do i=1, n_sources
       if(((usermagn .AND. rmagn(i-1) >= rmin .AND. rmagn(i-1) <= rmax) .OR. &
            .NOT.usermagn) .AND. &
          ((usebmagn .AND. bmagn(i-1) >= bmin .AND. bmagn(i-1) <= bmax) .OR. &
               .NOT.usebmagn)) then
          nnew = nnew + 1
          list_ptr(nnew)%dec = dec(i-1)*deg2rad 
       end if
      end do
    else if (index(units(decCol),"rad")>0) then
      nnew=0
      do i=1, n_sources
       if(((usermagn .AND. rmagn(i-1) >= rmin .AND. rmagn(i-1) <= rmax) .OR. &
            .NOT.usermagn) .AND. &
          ((usebmagn .AND. bmagn(i-1) >= bmin .AND. bmagn(i-1) <= bmax) .OR. &
               .NOT.usebmagn)) then
          nnew = nnew + 1
          list_ptr(nnew)%dec = dec(i-1)
       end if
      end do
    else
      call warning("NoUnit","No DEC unit; deg assumed")
      nnew=0
      do i=1, n_sources
       if(((usermagn .AND. rmagn(i-1) >= rmin .AND. rmagn(i-1) <= rmax) .OR. &
            .NOT.usermagn) .AND. &
          ((usebmagn .AND. bmagn(i-1) >= bmin .AND. bmagn(i-1) <= bmax) .OR. &
               .NOT.usebmagn)) then
          nnew = nnew + 1 
          list_ptr(nnew)%dec = dec(i-1)*deg2rad
       end if
      end do
    end if
       
    if (hasColumn(source_tab,cerr)) then

      errCol = column(source_tab, cerr, READ)
      err => real32Data( errCol )

      if (index(units(errCol),"sec")>0) then
        nnew=0
        do i=1, n_sources
         if(((usermagn .AND. rmagn(i-1) >= rmin .AND. &
              rmagn(i-1) <= rmax) .OR. .NOT.usermagn) .AND. &
            ((usebmagn .AND. bmagn(i-1) >= bmin .AND. &
              bmagn(i-1) <= bmax) .OR. .NOT.usebmagn)) then
            nnew = nnew + 1
!           NOTE: the CAT_RADEC_ERR is a 1 dimensional error.
            list_ptr(nnew)%error = err(i-1)
            list_ptr(nnew)%ra_err = err(i-1)
            list_ptr(nnew)%dec_err = err(i-1)
         end if
        end do
      else if (index(units(errCol),"deg")>0) then
        nnew=0
        do i=1, n_sources
         if(((usermagn .AND. rmagn(i-1) >= rmin .AND. &
              rmagn(i-1) <= rmax) .OR. .NOT.usermagn) .AND. &
            ((usebmagn .AND. bmagn(i-1) >= bmin .AND. &
              bmagn(i-1) <= bmax) .OR. .NOT.usebmagn)) then
            nnew = nnew + 1
            list_ptr(nnew)%error = err(i-1)*3600d0
            list_ptr(nnew)%ra_err = err(i-1)*3600d0
            list_ptr(nnew)%dec_err = err(i-1)*3600d0
         end if
        end do
      else if (index(units(errCol),"rad")>0) then
        nnew=0
        do i=1, n_sources
         if(((usermagn .AND. rmagn(i-1) >= rmin .AND. &
              rmagn(i-1) <= rmax) .OR. .NOT.usermagn) .AND. &
            ((usebmagn .AND. bmagn(i-1) >= bmin .AND. &
              bmagn(i-1) <= bmax) .OR. .NOT.usebmagn)) then
            nnew = nnew + 1
            list_ptr(nnew)%error = err(i-1)*3600d0/deg2rad
            list_ptr(nnew)%ra_err = err(i-1)*3600d0/deg2rad
            list_ptr(nnew)%dec_err = err(i-1)*3600d0/deg2rad
         end if
        end do
      else
        call warning("NoUnit","No error unit; arcsec assumed")
        nnew=0
        do i=1, n_sources
         if(((usermagn .AND. rmagn(i-1) >= rmin .AND. &
              rmagn(i-1) <= rmax) .OR. .NOT.usermagn) .AND. &
            ((usebmagn .AND. bmagn(i-1) >= bmin .AND. &
              bmagn(i-1) <= bmax) .OR. .NOT.usebmagn)) then
            nnew = nnew + 1
            list_ptr(nnew)%error = err(i-1)
            list_ptr(nnew)%ra_err = err(i-1)
            list_ptr(nnew)%dec_err = err(i-1)
         end if
        end do
      end if

    else
      nnew=0
      do i=1, n_sources
        if(((usermagn .AND. rmagn(i-1) >= rmin .AND. &
              rmagn(i-1) <= rmax) .OR. .NOT.usermagn) .AND. &
           ((usebmagn .AND. bmagn(i-1) >= bmin .AND. &
              bmagn(i-1) <= bmax) .OR. .NOT.usebmagn)) then
          nnew = nnew + 1
          list_ptr(nnew)%error = 1d0
          list_ptr(nnew)%ra_err = 1d0
          list_ptr(nnew)%dec_err = 1d0
        end if
      end do
    end if
    n_sources = nnew

    call release( raCol )
    call release( decCol )
    call release( errCol )
    if( usebmagn ) then 
       call release( bmagnCol )
    else 
       deallocate( bmagn)
    end if
    if( usermagn ) then 
       call release( rmagnCol )
    else 
       deallocate( rmagn)
    end if

    call release(source_tab)
    call release(set)


  end function read_catset

  !--------------------------------------------------------------------
  subroutine modify_list(filename, source_list, n_sources, &
       ra_shift, dec_shift, rotation, ra_sh_err, dec_sh_err, rot_err, &
       max_likelihood, n_matches)
  !--------------------------------------------------------------------

    character(len=*), intent(in) :: filename
    type(Coordinate_type), dimension(:) :: source_list
    integer(kind=int32), intent(in) :: n_sources, n_matches
    real(kind=double),intent(in) :: ra_shift, dec_shift, rotation
    real(kind=double),intent(in) :: ra_sh_err, dec_sh_err, rot_err
    real(kind=double),intent(in) :: max_likelihood

    type(DataSetT) :: set
    type(TableT)   :: source_tab
    type(ColumnT)  :: raCol, decCol, liiCol, biiCol
    integer(kind=int32) :: i
    real(kind=double),dimension(:),pointer :: ra, dec, lii, bii

    if( n_sources <= 0 ) then
       call error('noData','No data available for corrected positions!')
    end if

    set = dataSet(filename, MODIFY)
    source_tab = table(set,"SRCLIST")

    call setAttribute(source_tab,"RAOFFSET",ra_shift,"arcsec",&
         "RA correction applied")
    call setAttribute(source_tab,"DEOFFSET",dec_shift,"arcsec",&
         "DEC correction applied")
    call setAttribute(source_tab,"RAOFFERR",ra_sh_err,"arcsec",&
         "Error of RA correction")
    call setAttribute(source_tab,"DEOFFERR",dec_sh_err,"arcsec",&
         "Error of DEC correction")
    call setAttribute(source_tab,"ROT_CORR",rotation,"degrees",&
         "Rotation correction applied")
    call setAttribute(source_tab,"ROT_ERR",rot_err,"degrees",&
         "Error in rotation correction")
    call setAttribute(source_tab,"LIK_HOOD",max_likelihood,"none",&
         "statistic")
    call setAttribute(source_tab,"NMATCHES",n_matches,"number",&
         "Number of optical-X-ray source matches")

!    if( hasColumn(source_tab,"RA") ) then
!       raCol = column(source_tab,"RA",MODIFY)
!    else
!       raCol = addColumn(source_tab,"RA",Real64,"deg")
!    end if

!    if( hasColumn(source_tab,"DEC") ) then
!       decCol = column(source_tab,"DEC",MODIFY)
!    else
!       decCol = addColumn(source_tab,"DEC",Real64,"deg")
!    end if

    if( hasColumn(source_tab,"RA_CORR") ) then
       raCol = column(source_tab,"RA_CORR",MODIFY)
    else
       raCol = addColumn(source_tab,"RA_CORR",Real64,"deg")
    end if

    if( hasColumn(source_tab,"DEC_CORR") ) then
       decCol = column(source_tab,"DEC_CORR",MODIFY)
    else
       decCol = addColumn(source_tab,"DEC_CORR",Real64,"deg")
    end if

    if( hasColumn(source_tab,"LII_CORR") ) then
       liiCol = column(source_tab,"LII_CORR",MODIFY)
    else
       liiCol = addColumn(source_tab,"LII_CORR",Real64,"deg")
    end if
    if( hasColumn(source_tab,"BII_CORR") ) then
       biiCol = column(source_tab,"BII_CORR",MODIFY)
    else
       biiCol = addColumn(source_tab,"BII_CORR",Real64,"deg")
    end if


    ra => real64Data(raCol)
    dec => real64Data(decCol)
    lii => real64Data(liiCol)
    bii => real64Data(biiCol)

    do i=1,n_sources
       ra(i-1) = source_list(i)%ra   * rad2deg
       dec(i-1) = source_list(i)%dec * rad2deg
       lii(i-1) = source_list(i)%lii_corr
       bii(i-1) = source_list(i)%bii_corr
    end do


    call release(raCol)
    call release(decCol)
    call release(liiCol)
    call release(biiCol)

    call release(source_tab)
    call release(set)

  end subroutine modify_list


 !--------------------------------------------------------------------
  subroutine write_image( filename, lik_image, nx, ny )
 !--------------------------------------------------------------------
    character(len=*), intent(in) :: filename
    real(kind=double),dimension(:) :: lik_image
    integer(kind=int16), intent(in) :: nx, ny

    integer(kind=int32) :: i,j,k
    integer(kind=int32), dimension(2) :: shape
    real(kind=double),dimension(:,:), pointer :: im
    type(ArrayT)   :: arr
    type(DataSetT) :: set

    set = dataSet(filename,CREATE)

    call addComment(set,"This image displays the grid with likelihood values.")
    call addComment(set,"The X/Y axes correspond to x/y offsets.")
    call addComment(set,"Intended for checking the optimization results by eye")

    shape(1) = nx
    shape(2) = ny
    arr = addArray(set,"PRIMARY",Real64,shape)
    im => real64Array2Data(arr)
    k=1
    do j=0,ny-1
       do i=0,nx-1
          im(i, j) = lik_image(k)
          k = k + 1
       end do
    end do

    call release (arr)
    call release (set)

  end subroutine write_image

end module eposcorr_io



