 !+
! Module: eposcorr_mod
!
! Authors:  Jacco Vink - AIP
!           Dieter Meinert - AIP (D.M.)
!          
! Description: Correction of EPIC source positions by cross-correlation with optical source catalogue.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eposcorr_mod

  use dal
  use param  
  use errorHandling
  use eposcorr_def
  use eposcorr_io
  use eposcorr_coord
  use eposcorr_sel


  implicit none 

  private likelihood, find_offset, find_rotation, find_all, find_rot_err,&
       listparameters, printresults

  integer(kind=int16),parameter :: search_size=15, rotsize=25, coursegridsize=5
  integer(kind=int16),parameter :: ngrid = 2*search_size+1
  real(kind=double),parameter :: min_likelihood=0.1

! introduce parameter maxposnerr D.M.20000919
! error of xray source position allowed for source to be included in 
! position fitting, dimension: arcsec
  real(kind=double) :: maxposnerr = 20.0 !ref maxdist
!

  logical           :: usebmagn, usermagn, usemaxdist, makeimage, findrotation, matchtab
  character(81)     :: xray_file, optical_file, image_file
  integer(kind=int16):: niterations=2
  integer(kind=int32):: n_matches=0
  real(kind=single) :: bmin=-10.0, bmax=30.0, rmin=-10.0, rmax=30.0
  real(kind=single) :: maxdist = 20.0
  real(kind=double) :: maximum_likelihood=0.0, total_mu=0.0, variance=0.0
  real(kind=double) :: gridspacing=1.0, deltarota=0.2
  real(kind=double) :: maxrotation=1.0, maxoffset=8.0

  real(kind=double) :: raoffset=0.0, decoffset=0.0, rotation=0.0
  real(kind=double) :: raoffset_err=0.0, decoffset_err=0.0, rot_err=0.0
  real(kind=double) :: ra_pnt=0.0, dec_pnt=0.0, rapole, decpole

  real(kind=double),dimension(ngrid*ngrid) :: likimage

   integer(kind=int32) :: nxray=0, noptical=0
  type(Coordinate_type), dimension(:), pointer :: xraylist
  type(Coordinate_type), dimension(:), pointer :: opticallist

  character(160)    :: messagestr

  public eposcorr



contains 
  
  !---------------------------------------------------------------------
  subroutine eposcorr  
  !---------------------------------------------------------------------
    integer(kind=int16) :: i
    integer(kind=int32) :: nm
    real(kind=single) :: averror, lik
    character(len=20) :: xra, xdec, xerr, xext, cra,cdec,cerr,cbmg,crmg,cext
  
    xray_file = stringParameter('xrayset')
    optical_file = stringParameter('opticalset')
    niterations=intParameter('niter')
    findrotation = booleanParameter('findrotation')
    maxrotation = realParameter('maxrotation')
    maxoffset   = realParameter('maxoffset')
    maxdist =  realParameter('maxdist')
    usemaxdist =  booleanParameter('usemaxdist')
    usebmagn =  booleanParameter('usebmagn')
    bmin = realParameter('minbmagn')
    bmax = realParameter('maxbmagn')
    usermagn =  booleanParameter('usermagn')
    rmin = realParameter('minrmagn')
    rmax = realParameter('maxrmagn')
    makeimage = booleanParameter('makeimage')
    image_file = stringParameter('imagename')
    maxposnerr = realParameter('maxposnerr')
    xra = stringParameter('xrayra')
    xdec = stringParameter('xraydec')
    xerr = stringParameter ('xrayradecerr')
    xext = stringParameter ('xraytableext')
    cra = stringParameter('opticalra')
    cdec = stringParameter('opticaldec')
    cerr = stringParameter ('opticalradecerr')
    cbmg = stringParameter('bmag')
    crmg = stringParameter('rmag')
    cext = stringParameter ('opticaltableext')
    matchtab=booleanParameter("withmatchtable")
      
    call listparameters

!
!   Read the input data sets
!
    xraylist => read_srclist( xray_file, nxray, ra_pnt, dec_pnt, &
                              xra, xdec, xerr, xext )
    opticallist => read_catset( optical_file, noptical, bmin, bmax, &
         usebmagn, rmin, rmax, usermagn, cra, cdec, cerr, cbmg, crmg, cext )
    if( abs(dec_pnt - nodec_pnt) < 0.1 ) then
       call warning('noRaDecPnt','Keywords ra_pnt/dec_pnt not found')
       findrotation = .FALSE.
    end if

    if ( nxray > 0 .AND. noptical > 0 ) then

     write(errstr,*) 'Data sets were succesfully read. Processing...'
     call message(AppMsg,Verbose)

     if (booleanParameter("calculateoffsets")) then 
       
       if( usemaxdist ) then
          call message(AppMsg,Verbose,&
               'Minimizing the size of the optical data set.')
          write(errstr,*) 'Changing from   noptical = ', noptical
          call message(AppMsg,Verbose)

          call preselectlist(xraylist, nxray, opticallist, noptical, maxdist )

          write(errstr,*) '           to   noptical = ', noptical
          call message(AppMsg,Verbose)
       end if

       averror = meanerror(xraylist, nxray)
!       PRINT "(A F8.3)", 'average error = ', averror
       
       deltarota  = maxrotation/coursegridsize
       gridspacing = maxoffset/coursegridsize


       if( findrotation) then
          write(errstr,'(A,F8.3,A,F8.3)') 'Using: offset stepsize = ',&
               gridspacing,', rotation stepsize = ',deltarota
          call message(AppMsg,Verbose)

          call find_all

          call message(AppMsg,Verbose,' Results of the course 3D grid search:')

          call printresults
          gridspacing = 0.33*gridspacing 
          deltarota = 5.0*deltarota/(rotsize+1)
       end if

       do i=1, niterations
          write(errstr,*) 'Iteration nr ',i
          call message(AppMsg,Verbose)

          write(errstr,'(A,F8.3,A,F8.3)') 'Using: offset stepsize = ',&
               gridspacing,', rotation stepsize = ',deltarota
          call message(AppMsg,Verbose)
               
          if( i /= niterations) then
             call find_offset
             raoffset_err = 0.0
             decoffset_err = 0.0
             if( findrotation ) call find_rotation
          else
             gridspacing = 7.0*averror/search_size
             call find_offset
             if( findrotation ) then
                deltarota = 0.45*averror/search_size
                call find_rotation
                call find_rot_err
             end if
             if( makeimage ) then
                call write_image(image_file, likimage, ngrid, ngrid)
                makeimage = .FALSE.
             end if
          end if

          call printresults

          if( i == niterations-1) then

          else
             gridspacing = 0.33*gridspacing 
             deltarota = 0.33*deltarota
          end if
       end do
       

       if( n_matches <= n_min_matches .OR. &
            maximum_likelihood <= min_likelihood ) then
          call warning('invalidLikelihood', &
               'The offsets found should be considered unreliable')
       end if

       call message(AppMsg,Verbose,'Final result:')
       call printresults
       if (matchtab) lik=likelihood(raoffset,decoffset,rotation, nm, .true.)

     else
       raoffset=realParameter("raoffset")
       decoffset=realParameter("decoffset")
       rotation=realParameter("rotation")
     end if

     call correct_positions( xraylist, nxray, &
            ra_pnt*deg2rad, dec_pnt*deg2rad, raoffset, decoffset, rotation)
       

     call calc_gal_coord( xraylist, nxray)

     call modify_list(xray_file, xraylist, nxray, raoffset, decoffset, &
          rotation, raoffset_err, decoffset_err, rot_err,maximum_likelihood,&
            n_matches )
 
    else
       call error('emptyLists','One of the source lists is empty')
    end if

    if( nxray > 0 ) then
       deallocate( xraylist )
    endif

    if( noptical > 0 ) then
       deallocate( opticallist )
    endif

  end subroutine eposcorr

  !---------------------------------------------------------------------
  subroutine listparameters
  !---------------------------------------------------------------------

    write(errstr,*) 'Data set with X-ray source positions :',xray_file
    call message(AppMsg, Sparse)
    write(errstr,*) 'Data set with optical source positions :',optical_file
    call message(AppMsg, Noisy)
    if( usebmagn ) then
       write(errstr,*) 'B magn range : ',bmin,' - ', bmax
       call message(AppMsg,Verbose)
    end if
    if( usermagn ) then
       write(errstr,*) 'R magn range : ',rmin,' - ', rmax
       call message(AppMsg,Verbose)
    end if
    if( usemaxdist ) then
       write(errstr,*) &
            'Maximum distance of optical to nearest X-ray source = ',maxdist
       call message(AppMsg,Verbose)
    endif
    if( makeimage ) then
       write(errstr,*) 'An image will be made with file name :', image_file
       call message(AppMsg,Verbose)
    end if
    write(errstr,'(A,F8.2)') 'maximum rotation = ',maxrotation
    call message(AppMsg,Verbose)
    write(errstr,'(A,F8.2)') 'maximum offsets  = ',maxoffset
    call message(AppMsg,Verbose)


  end subroutine listparameters

  !---------------------------------------------------------------------
  subroutine printresults
  !---------------------------------------------------------------------

       write(errstr,'(A,F8.3,A,F8.3)') ' raoffset = ',raoffset,&
            ' +/- ',raoffset_err
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3,A,F8.3)') ' decoffset= ',decoffset,&
            ' +/- ',decoffset_err
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3,A,F8.3)') ' rotation = ',rotation, &
            ' +/- ',rot_err
       call message(AppMsg,Verbose)

       write(errstr,'(A,F8.3,A,I4)') ' maxlik   = ',&
            maximum_likelihood,' nm =', n_matches
       call message(AppMsg,Verbose)

  end subroutine printresults

  !---------------------------------------------------------------------
  subroutine find_all
  !---------------------------------------------------------------------
    integer(kind=int16) :: i, j, k
    integer(kind=int16) :: rotsize=coursegridsize, offsetsize=coursegridsize
    integer(kind=int32) :: nm,nmmax
    real(kind=double)   :: best_dra, best_ddec, bestrota,maxlik
    real(kind=double)   :: delta_ra, delta_dec, rota, lik

    maxlik = maximum_likelihood
    best_dra = raoffset
    best_ddec = decoffset
    bestrota = rotation
    nmmax= n_matches

    do k = -rotsize, rotsize

       rota = rotation + k * deltarota
       do j= -offsetsize, offsetsize
          delta_dec = decoffset + j*gridspacing
          do i = -offsetsize, offsetsize
             delta_ra = raoffset + i*gridspacing
             lik =  likelihood(delta_ra, delta_dec, rota, nm, .false.)
             if ( maxlik < lik ) then
                bestrota  = rota
                best_dra  = delta_ra
                best_ddec = delta_dec
                maxlik = lik
                nmmax = nm
             end if

          end do
       end do
    end do

    maximum_likelihood = maxlik
    n_matches = nmmax
    raoffset  = best_dra
    decoffset = best_ddec
    rotation  = bestrota
    !if (matchtab) lik=likelihood(best_dra, best_ddec, bestrota, nm, .true.)

  end subroutine find_all


  !---------------------------------------------------------------------
  subroutine find_rotation
  !---------------------------------------------------------------------

    integer(kind=int16) :: i
    integer(kind=int32) :: nm, nmmax
    real(kind=double)   :: bestrota, rota, lik, maxlik



    bestrota = rotation
    maxlik = likelihood(raoffset, decoffset, rotation, nm, .false. )
    nmmax = n_matches

    do i = -rotsize, rotsize

       rota = rotation + i * deltarota
       lik =  likelihood(raoffset, decoffset, rota, nm, .false.)
       if ( maxlik < lik ) then
          bestrota = rota
          maxlik = lik
          nmmax = nm
       end if

    end do

    if( maxlik > 0.0 ) then
       rotation = bestrota
       n_matches = nmmax
       !if (matchtab) lik=likelihood(raoffset, decoffset, bestrota, nm, .true.)
    end if
    
  end subroutine find_rotation

  !---------------------------------------------------------------------
  subroutine find_offset
  !---------------------------------------------------------------------

    integer(kind=int16) :: i, j, imax, jmax, k
    integer(kind=int32) :: nm, nmmax=0
    real(kind=double) :: delta_ra, delta_dec
    real(kind=double) :: raoffset0, decoffset0
    real(kind=double) :: best_dra, best_ddec, maxlik, lik, tot_lik
    real(kind=double) :: xm, ym, rm2, r, r2, x2, y2, r_max


    raoffset0 = raoffset
    decoffset0 = decoffset


    ! search for the optimum offset using a grid
    best_dra = raoffset0
    best_ddec = decoffset0
    nmmax = n_matches
    maxlik = maximum_likelihood
    imax=0
    jmax=0
    k = 0
    do j = -search_size, search_size
       
       delta_dec = decoffset0 + gridspacing * j
       do i = -search_size, search_size
          
          delta_ra = raoffset0 + gridspacing * i
          lik =  likelihood(delta_ra, delta_dec, rotation, nm, .false.)

          ! store the likelihood values:
          k = k + 1                            ! index for the likelihood grid
          likimage( k ) = lik
          
          if ( maxlik < lik ) then
             best_dra = delta_ra
             best_ddec = delta_dec
             maxlik = lik
             nmmax = nm
             imax = i + search_size
             jmax = j + search_size
          end if

       end do
    end do

    raoffset = best_dra        
    decoffset = best_ddec
    n_matches = nmmax
    maximum_likelihood = maxlik
    !if (matchtab) lik=likelihood(best_dra,best_ddec,rotation, nm, .true.) 
    
    if( n_matches <= 0 .OR. maximum_likelihood <= 0.0 ) then
!       call warning('noMatches',&
!            'No X-ray source seems to match optical source')
       return
    endif

    r_max = 0.8*gridspacing*search_size
    xm = 0.0
    ym = 0.0
    x2 = 0.0
    y2 = 0.0
    rm2 = 0.0
    tot_lik = 0.0
    k = 0
    do j = - search_size, search_size

       delta_dec = decoffset0 + gridspacing * j
       do i = -search_size, search_size

          k = k + 1
          delta_ra = raoffset0 + gridspacing * i
          r2 = (delta_ra - best_dra)**2 +  (delta_dec - best_ddec)**2
          r = sqrt( r2 )
          if( r <= r_max ) then
             lik = likimage(k)
             xm  = xm + lik * (delta_ra - best_dra);
             ym  = ym + lik * (delta_dec - best_ddec);
             x2  = x2 + lik * (delta_ra - best_dra)**2;
             y2  = y2 + lik * (delta_dec - best_ddec)**2;
             rm2  = rm2 + lik * r2
             tot_lik = tot_lik + lik; 
          end if

       end do

    end do


    if( tot_lik > 0.0 ) then

       ! x/decoffset are initially 0.0
       ! but are possibly /= 0.0 

!       raoffset = xm/tot_lik  + best_dra        
!       decoffset = ym/tot_lik + best_ddec
!       variance = rm2/tot_lik
!       maximum_likelihood = likelihood(raoffset,decoffset, rotation,n_matches )
!       PRINT "(A F8.3 F8.3 A F8.3 I4)",' offsets : ', raoffset, decoffset, &
!            ' maxlik = ', maximum_likelihood, n_matches,' better? ',&
!            likelihood( xm/tot_lik  + best_dra, ym/tot_lik + best_ddec,&
!            rotation,nm )

       if( n_matches-2 > 0 ) then
          raoffset_err = sqrt(x2/(tot_lik*(n_matches-2)))
          decoffset_err = sqrt(y2/(tot_lik*(n_matches-2)))
       else
          raoffset_err = 0.0
          decoffset_err = 0.0
       endif
    end if

  end subroutine find_offset

  !---------------------------------------------------------------------
  subroutine find_rot_err
  !---------------------------------------------------------------------

    integer(kind=int16) :: i
    integer(kind=int32) :: nm
    real(kind=double)   :: rota, totlik, lik, totrot2

    totlik = 0.0
    totrot2 = 0.0

    do i = -rotsize, rotsize

       rota = rotation + i * deltarota
       lik =  likelihood(raoffset, decoffset, rota, nm, .false. )
       totrot2 = totrot2 + lik*(rota - rotation)**2
       totlik = totlik + lik
    end do

    if( totlik > 0.0 .AND. n_matches > 3) then
       rot_err = sqrt(totrot2/(totlik - n_matches -3)) 
    end if


  end subroutine find_rot_err

  !---------------------------------------------------------------------
  function likelihood( raoffset, decoffset, rota, nmatches, mtab)
  !---------------------------------------------------------------------

    !
    ! the likelihood is define here as:
    !                                  sum(exp( -0.5(x_i-x_j)^2/sigma^2)
    ! with x_i and x_j resp. the x-ray  and optical positions of sources 
    ! which coincide (are within 5 sigma of each other), 
    ! sigma is the position error.
    ! Note that in case the association is real and there are no systematic 
    ! errors the average likelihood is sqrt(1/2) = 0.71
    ! with a variance of sqrt(1/3)-1/2= 0.077
    !

    real(kind=double) :: likelihood
    real(kind=double), intent(in) :: raoffset, decoffset, rota
    integer(kind=int32), intent(out) :: nmatches
    logical, intent(in) :: mtab
    
    integer(kind=int32) :: i, j
    real(kind=double) :: err, xra, xdec, ra0, dec0
    real(kind=double) :: deltaRA, angdist, cosdist, sindec, cosdec
    real(kind=double) :: rota_rad, raoffs_rad, decoffs_rad

    integer(kind=int32), dimension(:), pointer :: match1,match2
    real(kind=single), dimension(:), pointer :: match3,match4,match5, &
                                  match6,match7,match8 
    integer(kind=int32), dimension(2,10000) :: imatchbuff
    real(kind=single), dimension(3:8,10000) :: matchbuff    


    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT), dimension(8) :: c

    !if (mtab) then
    !  allocate (match1(0:10000))
    !  allocate (match2(0:10000))
    !  allocate (match3(0:10000))
    !  allocate (match4(0:10000))
    !  allocate (match5(0:10000))
    !  allocate (match6(0:10000))
    !  allocate (match7(0:10000))
    !  allocate (match8(0:10000))
    !end if

    likelihood = 0.0
    nmatches = 0
    
    if( nxray <= 0 .OR. noptical <= 0 ) then
       call error('emptyLists','One of the source lists does not contain data')
       return
    end if

    ra0 = ra_pnt * deg2rad
    dec0 = dec_pnt * deg2rad
    rota_rad = rota * deg2rad
    raoffs_rad = raoffset * arcsec2rad
    decoffs_rad = decoffset * arcsec2rad

    do i=1, noptical
       cosdec = dcos( opticallist(i)%dec )
       sindec = dsin( opticallist(i)%dec )

       do j=1,nxray
! Add parameter maxposnerr here D.M.20000919
          if ( xraylist(j)%ra_err <= maxposnerr .AND. &
               xraylist(j)%dec_err <= maxposnerr ) then
! 
             xra = xraylist(j)%ra
             xdec = xraylist(j)%dec
             call coordinate_transform(xra, xdec, ra0, dec0, pi, xra, xdec)
             
             !
             ! perform rotation
             !
             xra = xra +  rota_rad
             !
             ! transform backward
             !
             call coordinate_transform(xra, xdec, pi, dec0, ra0, xra, xdec )
             
             err = sqrt( opticallist(i)%ra_err**2 + opticallist(i)%dec_err**2 +&
                  xraylist(j)%ra_err**2 + xraylist(j)%dec_err**2 )*arcsec2rad
             
             if ( err /= 0.0 ) then
                
                if ( abs(xdec-opticallist(i)%dec-decoffs_rad)/err < max_error) &
                     then
                   if( abs(cosdec) > 1.0e-8 ) then
                      deltaRA = (xra - opticallist(i)%ra - raoffs_rad/cosdec)
                   else
                      deltaRA = (xra - opticallist(i)%ra)
                   end if
                   
                   cosdist = dsin(xdec - decoffs_rad) * sindec + &
                        dcos(xdec - decoffs_rad) * cosdec * dcos(deltaRA)
                   
                   if( cosdist > 0.99 ) cosdist = min(1.0d0, cosdist) 
                   if( cosdist < -0.99 ) cosdist = max(-1.0d0, cosdist) 
                   
                   angdist = dacos(cosdist)/err             
                   
                   if( angdist < max_error ) then
                      likelihood= likelihood+exp( -0.5*angdist*angdist)
                      nmatches = nmatches + 1
                      if (mtab .and. nmatches <=10000) then 
                        imatchbuff(1,nmatches)=j
                        imatchbuff(2,nmatches)=i
                        matchbuff(3,nmatches)=xra/deg2rad
                        matchbuff(4,nmatches)=xdec/deg2rad
                        matchbuff(5,nmatches)=opticallist(i)%ra/deg2rad
                        matchbuff(6,nmatches)=opticallist(i)%dec/deg2rad
                        matchbuff(7,nmatches)=dacos(cosdist)/deg2rad*3600.
                        matchbuff(8,nmatches)=exp(-0.5*angdist**2)
                      end if 

                   end if
                   
                end if
             end if
! end if parameter maxposnerr D.M.20000919
          end if
!
       end do
    end do

    if (mtab) then
      if (nmatches>10000) call warning("BufferFull", &
         "Buffer full; only first 10000 matches listed")

      set = dataSet(stringParameter("matchset"),CREATE)
      tab = addTable (set, "SRCLIST", nmatches)
      c(1)=addColumn(tab,"X-INDEX",Integer32," ")
      match1 => int32Data(c(1))
      match1 = imatchbuff(1,1:nmatches)
      c(3)=addColumn(tab,"X-RA",Real32,"deg")
      match3 => real32Data(c(3))
      match3 = matchbuff(3,1:nmatches)
      c(4)=addColumn(tab,"X-DEC",Real32,"deg")
      match4 => real32Data(c(4))
      match4 = matchbuff(4,1:nmatches)
      c(2)=addColumn(tab,"O-INDEX",Integer32," ")
      match2 => int32Data(c(2))
      match2 = imatchbuff(2,1:nmatches)
      c(5)=addColumn(tab,"O-RA",Real32,"deg")
      match5 => real32Data(c(5))
      match5 = matchbuff(5,1:nmatches)
      c(6)=addColumn(tab,"O-DEC",real32,"deg")
      match6 => real32Data(c(6))
      match6 = matchbuff(6,1:nmatches)
      c(7)=addColumn(tab,"DIST",real32,"arcsec")
      match7 => real32Data(c(7))
      match7 = matchbuff(7,1:nmatches)
      c(8)=addColumn(tab,"LIKELIHOOD",real32," ")
      match8 => real32Data(c(8))
      match8 = matchbuff(8,1:nmatches)
      do i=1,8
        call release(c(i))
      end do
      call release(tab)
      call release(set)

    end if

  end function likelihood


end module eposcorr_mod
